@@ -1,11 +1,10 @@
-/* Copyright (C) 1998, 1999 Thorsten Kukuk
+/* Copyright (C) 1998, 1999, 2000, 2001, 2002 Thorsten Kukuk
    This file is part of the yp-tools.
    Author: Thorsten Kukuk <kukuk@suse.de>
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2, or (at your option)
-   any later version.
+   it under the terms of the GNU General Public License version 2 as
+   published by the Free Software Foundation.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -21,8 +20,17 @@
 #include "config.h"
 #endif
 
+#define _GNU_SOURCE
+
+#define keydat_val dptr
+#define keydat_len dsize
+#define valdat_val dptr
+#define valdat_len dsize
+
 #include <pwd.h>
-#ifdef HAVE_CRYPT_H
+#ifdef HAVE_XCRYPT_H
+#include <xcrypt.h>
+#elif defined (HAVE_CRYPT_H)
 #include <crypt.h>
 #endif
 #include <ctype.h>
@@ -31,13 +39,22 @@
 #else
 #include "lib/getopt.h"
 #endif
+#include <errno.h>
+#include <time.h>
+#include <netdb.h>
+#include <limits.h>
 #include <string.h>
 #include <unistd.h>
+#include <stdlib.h>
 #include <locale.h>
 #include <libintl.h>
 #include <sys/param.h>
+#include <sys/stat.h>
 #include <rpcsvc/yp_prot.h>
 #include <rpcsvc/ypclnt.h>
+#ifdef HAVE_RPC_CLNT_SOC_H
+#include <rpc/clnt_soc.h>
+#endif
 
 #include "yppasswd.h"
 
@@ -52,6 +69,7 @@ int getrpcport(char *host, int prognum, int versnum, int proto);
 
 /* ok, we're using the crack library */
 #ifdef USE_CRACKLIB
+#include <crack.h>
 #ifndef CRACKLIB_DICTPATH
 #define CRACKLIB_DICTPATH "/usr/lib/cracklib_dict"
 #endif
@@ -157,14 +175,16 @@ getnismaster (char *domainname, const char *progname)
   port = getrpcport (master, YPPASSWDPROG, YPPASSWDPROC_UPDATE, IPPROTO_UDP);
   if (port == 0)
     {
-      fprintf (stderr, _("%s: yppasswdd not running on NIS master host\n"),
-               progname);
+      fprintf (stderr,
+	       _("%s: yppasswdd not running on NIS master host (\"%s\").\n"),
+               progname, master);
       return NULL;
     }
   if (port >= IPPORT_RESERVED)
     {
-      fprintf (stderr, _("%s: yppasswd daemon running on illegal port.\n"),
-               progname);
+      fprintf (stderr,
+	       _("%s: yppasswd daemon running on illegal port (\"%s\").\n"),
+               progname, master);
       return NULL;
     }
 
@@ -227,7 +247,7 @@ ypgetpw (char *master, char *domainname, char *name, int uid)
   struct ypreq_key key;
   enum clnt_stat res;
   struct ypresp_val resp;
-  char buffer[1024], uidbuf[8], *ptr, *keyval;
+  char *buffer, uidbuf[256], *ptr, *keyval;
   static struct passwd pwd;
   char *map;
   CLIENT *clnt;
@@ -237,7 +257,9 @@ ypgetpw (char *master, char *domainname, char *name, int uid)
 
   if (name == NULL)
     {
-      sprintf (uidbuf, "%d", uid);
+      if (snprintf (uidbuf, sizeof (uidbuf), "%d", uid) >= sizeof (uidbuf))
+	abort ();
+
       keyval = strdup (uidbuf);
       map = strdup ("passwd.byuid");
     }
@@ -246,6 +268,7 @@ ypgetpw (char *master, char *domainname, char *name, int uid)
       keyval = strdup (name);
       map = strdup ("passwd.byname");
     }
+
   /* Fill in values. */
   memset (&resp, 0, sizeof (struct ypresp_val));
   key.domain = strdup (domainname);
@@ -272,6 +295,7 @@ ypgetpw (char *master, char *domainname, char *name, int uid)
       return NULL;
     }
 
+  buffer = alloca (resp.valdat.valdat_len + 1);
   strncpy (buffer, resp.valdat.valdat_val, resp.valdat.valdat_len);
   buffer[resp.valdat.valdat_len] = '\0';
   free (resp.valdat.valdat_val);
@@ -344,11 +368,10 @@ getfield (char *gecos, char *field, int size)
   return sp;
 }
 
-#ifndef USE_CRACKLIB
-
+#if ! defined(USE_CRACKLIB) || defined(USE_CRACKLIB_STRICT)
 /* this function will verify the user's password
  * for some silly things.  If we're using cracklib, then
- * this function is unused.
+ * this function is unused unless USE_CRACKLIB_STRICT is set.
  */
 
 static int /* return values: 0 = not ok, 1 = ok */
@@ -366,9 +389,9 @@ verifypassword (struct passwd *pwd, char *pwdstr, uid_t uid)
   other = ucase = lcase = 0;
   for (p = pwdstr; *p; p++)
     {
-      ucase = ucase || isupper (*p);
-      lcase = lcase || islower (*p);
-      other = other || !isalpha (*p);
+      ucase = ucase || isupper ((int)*p);
+      lcase = lcase || islower ((int)*p);
+      other = other || !isalpha ((int)*p);
     }
 
   if ((!ucase || !lcase) && !other && uid)
@@ -424,6 +447,7 @@ main (int argc, char **argv)
   uid_t uid;
 
   setlocale (LC_MESSAGES, "");
+  setlocale (LC_CTYPE, "");
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
@@ -432,6 +456,13 @@ main (int argc, char **argv)
   else
     progname = argv[0];
 
+  if ((strcmp (progname, "ypchsh") == 0) ||
+      (strcmp (progname, "chsh") == 0))
+    l_flag = 1;
+  else if ((strcmp (progname, "ypchfn") == 0) ||
+	   (strcmp (progname, "chfn") == 0))
+    f_flag = 1;
+
   while (1)
     {
       int c;
@@ -445,7 +476,7 @@ main (int argc, char **argv)
       };
 
       c = getopt_long (argc, argv,
-		       (strcmp ("yppasswd", progname) == 0) ? "flp?" : "?",
+		       (l_flag == 0 && f_flag == 0) ? "flp?" : "?",
 		       long_options, &option_index);
       if (c == (-1))
         break;
@@ -461,9 +492,9 @@ main (int argc, char **argv)
 	  p_flag = 1;
 	  break;
 	case '?':
-	  if (strcmp (progname, "ypchsh") == 0)
+	  if (l_flag)
 	    print_help_chsh ();
-	  else if (strcmp (progname, "ypchfn") == 0)
+	  else if (f_flag)
 	    print_help_chfn ();
 	  else
 	    print_help_pwd ();
@@ -472,17 +503,17 @@ main (int argc, char **argv)
 	  print_version (progname);
 	  return 0;
 	case '\254':
-	  if (strcmp (progname, "ypchsh") == 0)
+	  if (l_flag)
 	    print_usage_chsh (stdout);
-	  else if (strcmp (progname, "ypchfn") == 0)
+	  else if (f_flag)
 	    print_usage_chfn (stdout);
 	  else
 	    print_usage_pwd (stdout);
 	  return 0;
 	default:
-	  if (strcmp (progname, "ypchsh") == 0)
+	  if (l_flag)
 	    print_usage_chsh (stderr);
-	  else if (strcmp (progname, "ypchfn") == 0)
+	  else if (f_flag)
 	    print_usage_chfn (stderr);
 	  else
 	    print_usage_pwd (stderr);
@@ -493,11 +524,6 @@ main (int argc, char **argv)
   argc -= optind;
   argv += optind;
 
-  if (strcmp (progname, "ypchsh") == 0)
-    l_flag = 1;
-  else if (strcmp (progname, "ypchfn") == 0)
-    f_flag = 1;
-
   if (f_flag == 0 && l_flag == 0)
     p_flag = 1;
 
@@ -509,6 +535,51 @@ main (int argc, char **argv)
   else if (argc == 1)
     user = argv[0];
 
+#ifdef YPPASSWD_IS_DEPRECATED
+  if (p_flag)
+    {
+      struct stat st;
+      if (stat ("/etc/SuSE-release", &st) == 0 &&
+	  stat (PASSWD_PROG, &st) == 0)
+	{
+	  printf (_("yppasswd is deprecated, use %s instead\n\n"),
+		  PASSWD_PROG);
+	  if (execl (PASSWD_PROG, "passwd", user, NULL) == -1)
+	    fprintf (stderr, _("Calling %s failed: %s\n"), PASSWD_PROG,
+		     strerror (errno));
+	}
+    }
+
+  if (l_flag)
+    {
+      struct stat st;
+      if (stat ("/etc/SuSE-release", &st) == 0 &&
+	  stat (CHSH_PROG, &st) == 0)
+	{
+	  printf (_("ypchsh is deprecated, use %s instead\n\n"),
+		  CHSH_PROG);
+	  if (execl (CHSH_PROG, "chsh", user, NULL) == -1)
+	    fprintf (stderr, _("Calling %s failed: %s\n"), CHSH_PROG,
+		     strerror (errno));
+	}
+    }
+
+  if (f_flag)
+    {
+      struct stat st;
+      if (stat ("/etc/SuSE-release", &st) == 0 &&
+	  stat (CHFN_PROG, &st) == 0)
+	{
+	  printf (_("ypchfn is deprecated, use %s instead\n\n"),
+		  CHFN_PROG);
+
+	  if (execl (CHFN_PROG, "chfn", user, NULL) == -1)
+	    fprintf (stderr, _("Calling %s failed: %s\n"), CHFN_PROG,
+		     strerror (errno));
+	}
+    }
+#endif
+
   if ((error = yp_get_default_domain (&domainname)) != 0)
     {
       fprintf (stderr, _("%s: can't get local yp domain: %s\n"),
@@ -524,7 +595,8 @@ main (int argc, char **argv)
   uid = getuid ();
   if ((pwd = ypgetpw (master, domainname, user, uid)) == NULL)
     {
-      fprintf (stderr, _("%s: unknown user (uid=%d).\n"), progname, uid);
+      fprintf (stderr, _("%s: unknown user (uid=%ld).\n"), progname,
+	       (long)uid);
       return 1;
     }
 
@@ -557,23 +629,47 @@ main (int argc, char **argv)
   if ((pwd->pw_passwd != NULL && strlen (pwd->pw_passwd) > 0) ||
       (pwd->pw_uid != uid))
     {
-      char prompt[40];
+      char prompt[130];
+      char *hashpass, *cp;
 
       if (pwd->pw_uid != uid)
-        sprintf (prompt, _("Please enter root password:"));
+        snprintf (prompt, sizeof (prompt), _("Please enter root password:"));
       else
-        sprintf (prompt, _("Please enter %spassword:"),
-		 p_flag ? _("old ") : "");
+        snprintf (prompt, sizeof (prompt), _("Please enter %spassword:"),
+		  p_flag ? _("old ") : "");
       s = getpass (prompt);
+      hashpass = alloca (strlen (pwd->pw_name) + 3);
+      cp = stpcpy (hashpass, "##");
+      strcpy (cp, pwd->pw_name);
 
       /* We can't check the password with shadow passwords enabled. We
        * leave the checking to yppasswdd */
-      if (uid != 0 && strcmp (pwd->pw_passwd, "x") != 0)
-        if (strcmp (crypt (s, pwd->pw_passwd), pwd->pw_passwd))
-          {
-            fprintf (stderr, _("Sorry.\n"));
-	    return 1;
-          }
+      if (uid != 0 && strcmp (pwd->pw_passwd, "x") != 0 &&
+	  strcmp (pwd->pw_passwd, hashpass ) != 0)
+	{
+	  int passwdlen;
+	  char *sane_passwd;
+	  passwdlen = strlen (pwd->pw_passwd);
+	  /* Some systems (HPU/X) store the password aging info after
+	   * the password (with a comma to separate it). To support
+	   * this we cut the password after the first invalid char
+	   * after the normal 13 ones. We can't cut at the first
+	   * invalid char, since MD5 uses $ in the first char.
+	   */
+	  if (passwdlen > 13)
+	    passwdlen = 13 + strspn(pwd->pw_passwd + 13,
+				    "abcdefghijklmnopqrstuvwxyz"
+				    "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
+				    "0123456789./");
+
+	  sane_passwd = alloca (passwdlen + 1);
+	  strncpy (sane_passwd, pwd->pw_passwd, passwdlen);
+	  if (strcmp (crypt (s, sane_passwd), sane_passwd))
+	    {
+	      fprintf (stderr, _("Sorry.\n"));
+	      return 1;
+	    }
+	}
       yppwd.oldpass = strdup (s);
     }
 
@@ -587,7 +683,7 @@ main (int argc, char **argv)
       int tries = 0;
       time_t tm;
 
-      buf = (char *) malloc (30);
+      buf = (char *) malloc (129);
 
       printf (_("Changing NIS password for %s on %s.\n"), pwd->pw_name,
 	      master);
@@ -609,7 +705,8 @@ main (int argc, char **argv)
 	      return 1;
 	    }
 
-	  strcpy (buf, p);
+	  strncpy (buf, p, 128);
+	  buf[128] = '\0';
 
 #ifdef USE_CRACKLIB
 	  error_msg = FascistCheck (buf, CRACKLIB_DICTPATH);
@@ -618,10 +715,12 @@ main (int argc, char **argv)
 	      fprintf (stderr, _("Not a valid password: %s.\n"), error_msg);
 	      continue;
 	    }
-#else
+#endif
+
+#if ! defined(USE_CRACKLIB) || defined(USE_CRACKLIB_STRICT)
 	  if (verifypassword (pwd, buf, uid) != 1)
 	    continue;
-#endif /* USE_CRACKLIB */
+#endif
 
 	  p = getpass (_("Please retype new password:"));
 	  if (strcmp (buf, p) == 0)
@@ -644,6 +743,7 @@ main (int argc, char **argv)
     {
       char gecos[1024], *sp, new_gecos[1024];
       char name[254], location[254], office[254], phone[254];
+      char oname[254], olocation[254], ooffice[254], ophone[254];
 
       printf (_("\nChanging full name for %s on %s.\n"
 	     "To accept the default, simply press return. To enter an empty\n"
@@ -651,20 +751,20 @@ main (int argc, char **argv)
 	      pwd->pw_name, master);
 
       strncpy (gecos, pwd->pw_gecos, sizeof (gecos));
-      sp = getfield (gecos, name, sizeof (name));
-      if (newfield (progname, _("Name"), strtok (gecos, ","), name,
+      sp = getfield (gecos, oname, sizeof (oname));
+      if (newfield (progname, _("Name"), oname, name,
 		    sizeof (name)))
 	return 1;
-      sp = getfield (sp, location, sizeof (location));
-      if (newfield (progname, _("Location"), location, location,
+      sp = getfield (sp, olocation, sizeof (olocation));
+      if (newfield (progname, _("Location"), olocation, location,
 		    sizeof (location)))
 	return 1;
-      sp = getfield (sp, office, sizeof (office));
-      if (newfield (progname, _("Office Phone"), office, office,
+      sp = getfield (sp, ooffice, sizeof (ooffice));
+      if (newfield (progname, _("Office Phone"), ooffice, office,
 		    sizeof (office)))
 	return 1;
-      sp = getfield (sp, phone, sizeof (phone));
-      if (newfield (progname, _("Home Phone"), phone, phone, sizeof (phone)))
+      sp = getfield (sp, ophone, sizeof (ophone));
+      if (newfield (progname, _("Home Phone"), ophone, phone, sizeof (phone)))
 	return 1;
       sprintf (new_gecos, "%s,%s,%s,%s", name, location, office, phone);
       sp = new_gecos + strlen (new_gecos);
@@ -725,9 +825,10 @@ main (int argc, char **argv)
 		 _("\nThe GECOS information has not been changed on %s.\n\n"),
 		   master);
 	}
-      fprintf (stdout,
-	       _("\nThe GECOS information has been changed on %s.\n\n"),
-	       master);
+      else
+	fprintf (stdout,
+		 _("\nThe GECOS information has been changed on %s.\n\n"),
+		 master);
       break;
     case 4:
       if (error || status)
@@ -740,8 +841,9 @@ main (int argc, char **argv)
 		   _("\nThe login shell has not been changed on %s.\n\n"),
 		   master);
 	}
-      fprintf (stdout,
-	       _("\nThe login shell has been changed on %s.\n\n"), master);
+      else
+	fprintf (stdout,
+		 _("\nThe login shell has been changed on %s.\n\n"), master);
       break;
     default:
       if (error || status)
@@ -755,9 +857,10 @@ main (int argc, char **argv)
 	   _("\nThe NIS account information has not been changed on %s.\n\n"),
 		   master);
 	}
-      fprintf (stdout,
-	       _("\nThe NIS account information has been changed on %s.\n\n"),
-	       master);
+      else
+	fprintf (stdout,
+		 _("\nThe NIS account information has been changed on %s.\n\n"),
+		 master);
       break;
     }
 
